/* petal.cc
 * 1997/09/09
 * (c)Cipher AKA Jeff Weeks of Code X
 *
 * This is the backbone of the PETAL system in action.  The code
 * here will load in a shared object as a driver, setup all function
 * pointers and you'll be ready for funky graphics.
 */

#ifdef __have_libdl__
#include <dlfcn.h>
#endif
#include <stdlib.h>
#include <iostream.h>

#include "petal.h"
#include "colour.h"
#include "lines.h"
#include "filled.h"
#include "poly.h"
#include "pixels.h"
#include "drivers/driver.h"

// The standard driver declaration.
Driver petal;
rgb pal[256];  // the default palette for 256 colour modes

// All the pointers to functions
int  (*colour)(float fr, float fg, float fb);
void (*putpixel)(int x, int y, int c, Image *img = petal.default_image);
int  (*getpixel)(int x, int y, Image *img = petal.default_image);
void (*fputpixel)(char *offset, int c);
int  (*fgetpixel)(char *offset);
void (*line) (int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image);
void (*hline)(int x1, int x2, int y, int c, Image *img = petal.default_image);
void (*vline)(int y1, int y2, int x, int c, Image *img = petal.default_image);
void (*clear)(int c, Image *img = petal.default_image);
void (*filled_rect)(int x1, int y1, int x2, int y2, int c, Image *img = petal.default_image);

/* The following function may be a bit confusing.  It really shouldn't be.
 * All it does is load in a shared library and extract vital functions that
 * will be used as the non-portable backend to PETAL
 */
int init_petal(char *drv_name, char *window_name) {
#ifdef __have_libdl__
  // Try to open the driver and report errors
  if(petal.handle) dlclose(petal.handle);
  petal.handle = dlopen(drv_name, RTLD_NOW);
  if(!petal.handle) {
    cout << "Couldn't open driver \"" << drv_name << "\"\n";
    return 1;
  }

  // Try to resolve the detect() function of the driver
  petal.detect = (int (*)(void)) dlsym(petal.handle, "detect");
  if(dlerror() != NULL) {
    cout << "Couldn't resolve \"int detect(void)\" function\n";
    dlclose(petal.handle);
    return 1;
  }

  // Check if the driver is able to be used (ie: we can't use the
  // xshm driver (X-Windows) in DOS.
  if(petal.detect()) {
    cout << "Invalid driver on current hardware\n";
    dlclose(petal.handle);
    return 1;
  }

  // At this point, the driver seems to be valid so we'll set up all the
  // function pointers of the petal structure.  We don't do any checking
  // at the moment either... if we've gotten this far, the driver should
  // be valid.
  petal.blit        = (void (*)(Image *, int, int))      dlsym(petal.handle, "blit");
  petal.get_image   = (void (*)(Image *, int, int))      dlsym(petal.handle, "get_image");
  petal.set_mode    = (int  (*)(char *, int, int, int))  dlsym(petal.handle, "set_mode");
  petal.deinit      = (void (*)(void))                   dlsym(petal.handle, "deinit");
  petal.get_pointer = (char (*)(int *,int *))            dlsym(petal.handle, "get_pointer");
  petal.get_key     = (char (*)(void))                   dlsym(petal.handle, "get_key");
  petal.set_pal     = (void (*)(rgb *))                  dlsym(petal.handle, "set_pal");

  // We store the window name for later use.  It'll be used when we set the
  // video mode, which, afterall, is where most window systems would want
  // this information.
  petal.window_name = window_name;
  return 0;
#else
  cout << "This system does not support dynamic loading\n";
  return 1;
#endif
}

int init_petal(Driver *drv, char *window_name) {
  // Check if the driver is able to be used (ie: we can't use the
  // xshm driver (X-Windows) in DOS.
  if(drv == NULL) return 1;
  if(drv->detect == NULL) return 1;
  if(drv->detect()) {
    cout << "Invalid driver on current hardware\n";
    return 1;
  }

  // At this point, the driver seems to be valid so we'll set up all the
  // function pointers of the petal structure.  We don't do any checking
  // at the moment either... if we've gotten this far, the driver should
  // be valid.
  petal.detect      = drv->detect;
  petal.blit        = drv->blit;
  petal.get_image   = drv->get_image;
  petal.set_mode    = drv->set_mode;
  petal.deinit      = drv->deinit;
  petal.get_pointer = drv->get_pointer;
  petal.get_key     = drv->get_key;
  petal.set_pal     = drv->set_pal;

  // We store the window name for later use.  It'll be used when we set the
  // video mode, which, afterall, is where most window systems would want
  // this information.
  petal.window_name = window_name;
  return 0;
}

int set_resolution(int x, int y, int bpp) {
  int status = petal.set_mode(petal.window_name,x,y,bpp);

  // We don't want to change anything unless the call was successful...
  if(!status) {
    switch(petal.display.bpp) {
      case 32: set_bpp32(); break;
      case 24: set_bpp24(); break;
      case 16: set_bpp16(); break;
      case 15: set_bpp15(); break;
      case  8: set_bpp8();  break;
    }

    // Now we'll have to allocate room for our scanline structures or
    // else we can't draw any polygons
    if(scanline) delete scanline;
    scanline = new Scanline[petal.display.height];
  }

  return status;
}

void deinit_petal(void) {
  if(petal.deinit) petal.deinit();
#ifdef __have_libdl__
  if(petal.handle) dlclose(petal.handle);
#endif
  return;
}

int get_resolution(int *x, int *y, int *bpp) {
  *x = petal.display.width;
  *y = petal.display.height;
  *bpp = petal.display.bpp;
  return petal.display.size;
}

Image *new_image(void) {
  Image *img = new Image;
  if(img == NULL) return NULL;

  // Allocate a virtual screen using the same specs as the display
  img->address = new char[petal.display.size];
  if(img->address == NULL) {
    delete img;
    return NULL;
  }
  img->size = petal.display.size;
  img->width = petal.display.width;
  img->height= petal.display.height;
  img->x_off = petal.display.x_off;
  img->y_off = petal.display.y_off;
  img->bytes_per_line = petal.display.bytes_per_line;
  img->bytes_per_pixel = petal.display.bytes_per_pixel;
  img->clipping = 0;
  img->transparent = 0;
  img->cx1 = img->cy1 = 0;
  img->cx2 = img->width;
  img->cy2 = img->height;
  img->bpp = petal.display.bpp;

  // Then return a pointer to the new screen
  return img;
}

Image *new_image(int x, int y) {
  Image *img = new Image;
  if(img == NULL) return NULL;

  // Allocate a virtual screen using the user supplied size
  img->width = x;
  img->height = y;
  img->clipping = 0;
  img->transparent = 0;
  img->cx1 = img->cy1 = 0;
  img->cx2 = img->width;
  img->cy2 = img->height;
  img->x_off = img->y_off = 0;
  img->bpp = petal.display.bpp;
  img->bytes_per_pixel = petal.display.bytes_per_pixel;
  img->bytes_per_line = x * img->bytes_per_pixel;
  img->size = img->height * img->bytes_per_line;
  img->address = new char[img->size];
  if(img->address == NULL) {
    delete img;
    return NULL;
  }

  // Then return a pointer to the new screen
  return img;
}

void delete_image(Image *img) {
  if(img->address) delete img->address;
  if(img) delete img;
}

void clip_image(int x, int y, int width, int height, Image *img = petal.default_image) {
  // We can't clip if it's out of range!
  if(x > img->width || y > img->height) return;

  img->cx1 = x;
  img->cy1 = y;

  if(x + width < img->width) img->cx2 = x + width;
  else img->cx2 = img->width;

  if(y + height < img->height) img->cy2 = y + height;
  else img->cy2 = img->height;

  img->clipping = true;
}

void get_image(Image *dest, Image *src, signed int x = 0, signed int y = 0) {
  int dsp_offset, img_offset, width, i;
  int pixel_width, pixel_height;
  int x1, y1, x2, y2, clipping;

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width or height
  // of the image goes off the screen
  clipping  = dest->clipping | src->clipping;
  clipping |= !(dest->size == src->size);
  clipping |= (dest->width  + x + src->x_off) > src->width;
  clipping |= (dest->height + y + src->y_off) > src->height;

  if(clipping) {
    // Perform some advanced clipping
    if(!clip(dest, src, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // Then a width in bytes
    width = pixel_width * dest->bytes_per_pixel;

    // Calculate starting offsets for the image and display (bottom left corner
    // so we can work from the bottom up)
    img_offset = (y1 +pixel_height) * dest->bytes_per_line + (x1 * dest->bytes_per_pixel);
    dsp_offset = (y + pixel_height) * src->bytes_per_line +  (x  * src->bytes_per_pixel);

    // copy clipped image to screen (work from bottom to top for speed)
    for(i = pixel_height; i; i--) {
      img_offset -= dest->bytes_per_line;
      dsp_offset -= src->bytes_per_line;
      memcpy(&dest->address[img_offset], &src->address[dsp_offset], width);
    }
  }
  else {
    // Copy the whole screen in one sweep if we can (no clipping needed)
    memcpy(dest->address, src->address, dest->size);
  }
  return;
}

void show_image(Image *img = petal.default_image, int x=0, int y=0) {
  if(img->clipping) {
    x += img->cx1;
    y += img->cy1;
  }
  petal.blit(img,x,y);
}

void show_image(int x, int y, int width, int height, Image *img = petal.default_image, int tx=0, int ty=0) {
  int cx1, cy1, cx2, cy2;
  int clipping;

  clipping = img->clipping;
  cx1 = img->cx1;
  cy1 = img->cy1;
  cx2 = img->cx2;
  cy2 = img->cy2;

  clip_image(x,y,width,height,img);

  //if(img->clipping) {
    tx += img->cx1;
    ty += img->cy1;
  //}


  /*
  if(img->cx1 + tx < petal.display.cx1) img->cx1 = petal.display.cx1 + tx;
  if(img->cy1 + ty < petal.display.cy1) img->cy1 = petal.display.cy1 + ty;
  if(img->cx2 + tx > petal.display.cx2) img->cx2 = petal.display.cx2 + tx;
  if(img->cy2 + ty > petal.display.cy2) img->cy2 = petal.display.cy2 + ty;
  */

  petal.blit(img, tx, ty);

  img->clipping = clipping;
  img->cx1 = cx1;
  img->cy1 = cy1;
  img->cx2 = cx2;
  img->cy2 = cy2;
}

void copy_image(int x, int y, int width, int height, Image *dest, Image *src = petal.default_image, int tx=0, int ty=0) {
  int cx1, cy1, cx2, cy2;
  int clipping;

  clipping = src->clipping;
  cx1 = src->cx1;
  cy1 = src->cy1;
  cx2 = src->cx2;
  cy2 = src->cy2;

  clip_image(x,y,width,height,src);

  /*
  if(src->cx1 + tx < dest->cx1) src->cx1 = dest->cx1 - tx;
  if(src->cy1 + ty < dest->cy1) src->cy1 = dest->cy1 - ty;
  if(src->cx2 + tx > dest->cx2) src->cx2 = dest->cx2 - tx;
  if(src->cy2 + ty > dest->cy2) src->cy2 = dest->cy2 - ty;
  */

  tx += src->cx1;
  ty += src->cy1;

  copy_image(dest, src, tx, ty);

  src->clipping = clipping;
  src->cx1 = cx1;
  src->cy1 = cy1;
  src->cx2 = cx2;
  src->cy2 = cy2;
}

void copy_image(Image *dest, Image *src = petal.default_image, int x = 0, int y = 0) {
  int dsp_offset, img_offset, width, i;
  int pixel_width, pixel_height;
  int x1, y1, x2, y2, clipping;
  int tx1, tx2, t, c;

  // We'll have to perform some type of clipping if either
  // window has their clipping flag set, or if the windows
  // are of different sizes.  Also if the width or height
  // of the image goes off the screen.  We'll also use the
  // clipping technique if the image is transparent.  This
  // isn't necessary, but it's easier and takes less code.
  clipping  = src->clipping | src->transparent | dest->clipping;
  clipping |= !(src->size == dest->size);
  clipping |= (src->width  + x + dest->x_off) > dest->width;
  clipping |= (src->height + y + dest->y_off) > dest->height;

  if(clipping) {
    // Clip the image (found in petal.h)
    if(!clip(src, dest, &x, &y, &x1, &y1, &x2, &y2)) return;

    // Calculate a width and height in pixels
    pixel_width  = (x2 - x1);
    pixel_height = (y2 - y1);

    // Then the width in bytes
    width = pixel_width * src->bytes_per_pixel;

    // Calculate starting offsets for the image and display (bottom left corner
    // so we can work from the bottom up)
    img_offset = (y1+ pixel_height) * src->bytes_per_line + (x1 * src->bytes_per_pixel);
    dsp_offset = (y + pixel_height) * dest->bytes_per_line + (x * dest->bytes_per_pixel);

    // copy clipped image to screen (work from bottom to top for speed)
    for(i = pixel_height; i; i--) {
      img_offset -= src->bytes_per_line;
      dsp_offset -= dest->bytes_per_line;

      // And display it (using either the transparent or full method)
      if(src->transparent) {
        // these will hold non-transparent spans to copy...
        tx1 = tx2 = x1 * src->bytes_per_pixel;
        t = src->trans_col;         // the transparent colour which will not be copied
        while(tx1 < width) {
          // Try and find the first non-transparent pixel
          c = fgetpixel(&src->address[img_offset + tx1]);
          if(c != t) {
            tx2 = tx1;  // start at the beginning of the non-transparent span
            // and try and find the end of it...
            while(fgetpixel(&src->address[img_offset + tx2]) != t &&
                  tx2 < width)
              tx2 += src->bytes_per_pixel;  // we haven't found the end yet so try the next pixel
            // Now draw the span
            memcpy(&dest->address[dsp_offset + tx1], &src->address[img_offset + tx1], tx2-tx1);
            tx1 = tx2;                     // search for the next span after this one, ofcourse
          }
          tx1 += src->bytes_per_pixel;    // goto the next pixel (haven't found start of span yet)
        }
      }
      else  // Not transparent, still clipped...
        memcpy(&dest->address[dsp_offset], &src->address[img_offset], width);
    }
  }
  else {
    // Copy the whole screen in one sweep if we can (no clipping needed)
    memcpy(dest->address, src->address, src->size);
  }
  return;
}


void transparent(int colour, Image *img = petal.default_image) {
  img->trans_col = colour;
  img->transparent = true;
}

void offset_image(int x, int y, Image *img = petal.default_image) {
  // Make sure we've got valid offsets
  if(x > img->width || y > img->height) return;
  // Adjust width and height to use the new offsets (taking into account
  // a possible previous offset_image call).  Why we do this is simple;
  // If we've offset the image, then the width and height must be changed
  img->width  = img->width  + img->x_off - x;
  img->height = img->height + img->y_off - y;
  img->x_off = x;
  img->y_off = y;
}

Image *sub_image(int x, int y, int width, int height, Image *img = petal.default_image) {
  Image *ret = new Image;

  // Check for validity
  if(x > img->width || y > img->height) return NULL;

  // Get a sensible width...
  if(x + width < img->width) ret->width = width;
  else ret->width = img->width - x;

  // ...and height
  if(y + height < img->height) ret->height = height;
  else ret->height = img->height - y;

  ret->clipping = 0;
  ret->transparent = img->transparent;
  ret->cx1 = ret->cy1 = 0;
  ret->cx2 = ret->width;
  ret->cy2 = ret->height;
  ret->x_off = ret->y_off = 0;
  ret->bpp = img->bpp;
  ret->bytes_per_pixel = img->bytes_per_pixel;
  ret->bytes_per_line  = img->bytes_per_line;
  ret->size = ret->height * ret->bytes_per_line;
  ret->address = img->address + y * img->bytes_per_line + x * img->bytes_per_pixel;

  // return a pointer to the sub image
  return ret;
}

void set_bpp8(void) {
  int i;

  max_col = 255;
  colour = colour8;
  putpixel = putpixel8;
  getpixel = getpixel8;
  fputpixel = fputpixel8;
  fgetpixel = fgetpixel8;
  line = line8;
  hline = hline8;
  vline = vline8;
  clear = clear8;
  filled_rect = filled_rect8;

  cout << "Setting default palette\n";
  for(i = 0; i < 256; i++) {
    pal[i].r = (i & 7) << 5;
    pal[i].g = (i & 56) << 3;
    pal[i].b = (i & 192);
  }

  set_pal(pal);
  return;
}

void set_bpp15(void) {
  max_col = 32767;
  colour = colour15;
  putpixel = putpixel16;
  getpixel = getpixel16;
  fputpixel = fputpixel16;
  fgetpixel = fgetpixel16;
  line = line16;
  hline = hline16;
  vline = vline16;
  clear = clear16;
  filled_rect = filled_rect16;
  return;
}

void set_bpp16(void) {
  max_col = 65535;
  colour = colour16;
  putpixel = putpixel16;
  getpixel = getpixel16;
  fputpixel = fputpixel16;
  fgetpixel = fgetpixel16;
  line = line16;
  hline = hline16;
  vline = vline16;
  clear = clear16;
  filled_rect = filled_rect16;
  return;
}

void set_bpp24(void) {
  max_col = 16777215;
  colour = colour24;
  putpixel = putpixel24;
  getpixel = getpixel24;
  fputpixel = fputpixel24;
  fgetpixel = fgetpixel24;
  line = line24;
  hline = hline24;
  vline = vline24;
  clear = clear24;
  filled_rect = filled_rect24;
  return;
}

void set_bpp32(void) {
  max_col = 0xffffffff;
  colour = colour32;
  putpixel = putpixel32;
  getpixel = getpixel32;
  fputpixel = fputpixel32;
  fgetpixel = fgetpixel32;
  line = line32;
  hline = hline32;
  vline = vline32;
  clear = clear32;
  filled_rect = filled_rect32;
  return;
}
